﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis.Interfaces.Classification;

namespace HeuristicLab.Problems.DataAnalysis {
  /// <summary>
  /// Base class for weight calculators for classification solutions in an ensemble.
  /// </summary>
  [StorableClass]
  public abstract class ClassificationWeightCalculator : NamedItem, IClassificationEnsembleSolutionWeightCalculator {
    [StorableConstructor]
    protected ClassificationWeightCalculator(bool deserializing) : base(deserializing) { }
    protected ClassificationWeightCalculator(ClassificationWeightCalculator original, Cloner cloner)
      : base(original, cloner) {
    }
    public ClassificationWeightCalculator()
      : base() {
      this.name = ItemName;
      this.description = ItemDescription;
    }

    private IEnumerable<double> weights;

    /// <summary>
    /// calls CalculateWeights and removes negative weights 
    /// </summary>
    /// <param name="classificationSolutions"></param>
    /// <returns>weights which are equal or bigger than zero</returns>
    public void CalculateNormalizedWeights(ItemCollection<IClassificationSolution> classificationSolutions) {
      List<double> weights = new List<double>();
      if (classificationSolutions.Count > 0) {
        foreach (var weight in CalculateWeights(classificationSolutions)) {
          weights.Add(weight >= 0 ? weight : 0);
        }
      }
      this.weights = weights.Select(x => x / weights.Sum());
    }

    protected abstract IEnumerable<double> CalculateWeights(ItemCollection<IClassificationSolution> classificationSolutions);

    public virtual IEnumerable<double> AggregateEstimatedClassValues(IEnumerable<IClassificationModel> models, Dataset dataset, IEnumerable<int> rows) {
      return from xs in ClassificationWeightCalculator.GetEstimatedClassValues(models, dataset, rows)
             select AggregateEstimatedClassValues(xs);
    }

    protected double AggregateEstimatedClassValues(IEnumerable<double> estimatedClassValues) {
      if (!estimatedClassValues.Count().Equals(weights.Count()))
        throw new ArgumentException("'estimatedClassValues' has " + estimatedClassValues.Count() + " elements, while 'weights' has" + weights.Count());
      IDictionary<double, double> weightSum = new Dictionary<double, double>();
      for (int i = 0; i < estimatedClassValues.Count(); i++) {
        if (!weightSum.ContainsKey(estimatedClassValues.ElementAt(i)))
          weightSum[estimatedClassValues.ElementAt(i)] = 0.0;
        weightSum[estimatedClassValues.ElementAt(i)] += weights.ElementAt(i);
      }
      if (weightSum.Count <= 0)
        return double.NaN;
      var max = weightSum.Max(x => x.Value);
      max = weightSum
        .Where(x => x.Value.Equals(max))
        .Select(x => x.Key)
        .First();
      return max;
    }

    protected static IEnumerable<IEnumerable<double>> GetEstimatedClassValues(IEnumerable<IClassificationModel> models, Dataset dataset, IEnumerable<int> rows) {
      if (!models.Any()) yield break;
      var estimatedValuesEnumerators = (from model in models
                                        select model.GetEstimatedClassValues(dataset, rows).GetEnumerator())
                                       .ToList();

      while (estimatedValuesEnumerators.All(en => en.MoveNext())) {
        yield return from enumerator in estimatedValuesEnumerators
                     select enumerator.Current;
      }
    }

    #region Helper
    protected IEnumerable<double> GetValues(IList<double> targetValues, IEnumerable<int> indizes) {
      return from i in indizes
             select targetValues[i];
    }
    #endregion
  }
}
