﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [StorableClass]
  [Item("MedianThresholdCalculator", "")]
  public class MedianThresholdCalculator : DiscriminantClassificationWeightCalculator {

    public MedianThresholdCalculator()
      : base() {
    }
    [StorableConstructor]
    protected MedianThresholdCalculator(bool deserializing) : base(deserializing) { }
    protected MedianThresholdCalculator(MedianThresholdCalculator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new MedianThresholdCalculator(this, cloner);
    }

    protected double[] threshold;
    protected double[] classValues;

    protected override IEnumerable<double> DiscriminantCalculateWeights(ItemCollection<IDiscriminantFunctionClassificationSolution> discriminantSolutions) {
      List<List<double>> estimatedTrainingValEnumerators = new List<List<double>>();
      List<List<double>> estimatedTrainingClassValEnumerators = new List<List<double>>();
      foreach (var solution in discriminantSolutions) {
        estimatedTrainingValEnumerators.Add(solution.EstimatedTrainingValues.ToList());
        estimatedTrainingClassValEnumerators.Add(solution.EstimatedTrainingClassValues.ToList());
      }

      List<double> median = new List<double>();

      IClassificationProblemData problemData = discriminantSolutions.ElementAt(0).ProblemData;
      List<double> targetValues = problemData.Dataset.GetDoubleValues(problemData.TargetVariable).ToList();
      IEnumerable<double> trainingVal = GetValues(targetValues, problemData.TrainingIndizes);

      for (int i = 0; i < estimatedTrainingClassValEnumerators.First().Count; i++) {
        var points = (from solution in estimatedTrainingValEnumerators
                      select solution[i])
                      .OrderBy(p => p)
                      .ToList();

        median.Add(GetMedian(points));
      }
      AccuracyMaximizationThresholdCalculator.CalculateThresholds(problemData, median, trainingVal, out classValues, out threshold);
      return median;
    }

    protected override double DiscriminantAggregateEstimatedClassValues(IEnumerable<double> estimatedClassValues, IEnumerable<double> estimatedValues) {
      double classValue = classValues.First();
      double median = GetMedian(estimatedValues.ToList());
      for (int i = 0; i < classValues.Count(); i++) {
        if (median > threshold[i])
          classValue = classValues[i];
        else
          break;
      }
      return classValue;
    }

    private double GetMedian(IList<double> estimatedValues) {
      int count = estimatedValues.Count;
      if (count % 2 == 0)
        return 0.5 * (estimatedValues[count / 2] + estimatedValues[count / 2 + 1]);
      else
        return estimatedValues[(count + 1) / 2];
    }
  }
}
